@@ -1,4 +1,4 @@
-__version__ = '0.58.1'
+__version__ = '0.61.2'
 
 from rdworks.conf    import Conf
 from rdworks.mol     import Mol
@@ -8,7 +8,7 @@ from rdworks.readin import read_csv, merge_csv, read_dataframe, read_smi, read_s
 from rdworks.std import desalt_smiles, standardize_smiles, standardize
 
 from rdworks.complete import complete_stereoisomers, complete_tautomers
-from rdworks.ionized import IonizedStates
+from rdworks.microstate import State, StateEnsemble, StateNetwork
 
 from rdworks.rgroup import expand_rgroup, most_common, most_common_in_NP
 from rdworks.scaffold import scaffold_network, scaffold_tree, BRICS_fragmented, BRICS_fragment_indices
