@@ -5,6 +5,7 @@ import numpy as np
 import ase
 import itertools
 
+from collections import namedtuple
 from collections.abc import Callable
 from typing import Self
 
@@ -205,7 +206,7 @@ class Conf:
         """
         if isinstance(calculator, str) :
             
-            PE_start = self.potential_energy(calculator)
+            PE_start = self.singlepoint(calculator)
             PE_final = None
             
             if calculator.lower() == 'xTB'.lower():
@@ -240,7 +241,7 @@ class Conf:
                 # returns 0 if the optimization converged
 
             if PE_final is None:
-                PE_final = self.potential_energy(calculator)
+                PE_final = self.singlepoint(calculator)
             
             self.props.update({
                 'E_tot_init(kcal/mol)': PE_start , # energy before optimization
@@ -421,7 +422,7 @@ class Conf:
         return np.sqrt(np.mean(b))
 
 
-    def potential_energy(self, calculator: str | Callable = 'MMFF94', **kwargs) -> float:
+    def singlepoint(self, calculator: str | Callable = 'MMFF94', **kwargs) -> float:
         """Get potential energy and set `E_tot(kcal/mol)` in the self.props.
 
         Args:
@@ -443,7 +444,6 @@ class Conf:
         Returns:
             float | None: potential energy in kcal/mol or None.
         """
-
         if isinstance(calculator, str):
             if calculator.lower() == 'xTB'.lower():
                 water = kwargs.get('water', None)
@@ -626,7 +626,7 @@ class Conf:
                 'Converged': [],
                 }
         
-        rdmols = []
+        confs = []
         if simplify:
             (frag, frag_ijkl, frag_created, wbo_filtered) = create_torsion_fragment(ref_conf.rdmol, indices)
             frag_conf = Conf(frag)
@@ -638,25 +638,28 @@ class Conf:
                 # to serialize the molecule
                 data['frag'] = Chem.MolToMolBlock(frag_cleaned)
                 data['frag_indices'] = frag_ijkl
-                rdmols.append(conf.rdmol)
                 data['angle'].append(angle)
+                confs.append(conf)
       
         else:
             for angle in np.arange(-180.0, 180.0, interval): 
                 # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
                 conf = ref_conf.copy()
                 conf.set_torsion(*indices, angle) # atoms bonded to `l` move.
-                rdmols.append(conf.rdmol)
                 data['angle'].append(angle)
+                confs.append(conf)
+                
+        batchsize_atoms = kwargs.get('batchsize_atoms', 16384)
+        batches = prep_confs_batches(confs, batchsize_atoms)
 
-        optimized = batchoptimizer(rdmols).run()
-
-        for rdmol in optimized.mols:
-            # conf.optimize() updates coordinates and conf.props: 
-            #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
-            data['init'].append(float(rdmol.GetProp('E_tot_init(kcal/mol)')))
-            data['last'].append(float(rdmol.GetProp('E_tot(kcal/mol)')))
-            data['Converged'].append(True if rdmol.GetProp('Converged') == 'True' else False)
+        for batch in batches:
+            optimized = batchoptimizer(batch.rdmols).run()
+            for rdmol in optimized.mols:
+                # conf.optimize() updates coordinates and conf.props: 
+                #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
+                data['init'].append(float(rdmol.GetProp('E_tot_init(kcal/mol)')))
+                data['last'].append(float(rdmol.GetProp('E_tot(kcal/mol)')))
+                data['Converged'].append(True if rdmol.GetProp('Converged') == 'True' else False)
 
         # Post-processing        
         if use_converged_only:
@@ -705,6 +708,8 @@ class Conf:
             fmax (float, optional): fmax of ASE optimizer. Defaults to 0.05.
             interval (float, optional): interval of torsion angles in degree. Defaults to 15.0.
             use_converged_only (bool, optional): whether to use only converged data. Defaults to True.
+            batchsize_atoms (int, optional): max number of atoms in a batch. 
+                Defaults to 16384(=16*1024). Disable batch optimization if smaller than conf.natoms.
 
         Returns:
             Self: modified self.
@@ -716,16 +721,29 @@ class Conf:
             # single torsion angle atom indices
 
         conf = self.copy()
-
         torsion_energy_profiles = {}
-        for tk, indices in torsion_atoms_dict.items():
-            conf = conf.torsion_energies_one(calculator, 
-                                             indices, 
-                                             simplify, 
-                                             fmax, 
-                                             interval, 
-                                             use_converged_only)
-            torsion_energy_profiles[tk] = conf.props['torsion']
+        if 'batchsize_atoms' in kwargs and kwargs.get('batchsize_atoms') >= conf.natoms:
+            batchsize_atoms = kwargs.get('batchsize_atoms')
+            for tk, indices in torsion_atoms_dict.items():
+                conf = conf.torsion_energies_batch(calculator, 
+                                                indices, 
+                                                simplify, 
+                                                fmax, 
+                                                interval, 
+                                                use_converged_only,
+                                                batchsize_atoms=batchsize_atoms,
+                                                )
+                torsion_energy_profiles[tk] = conf.props['torsion']
+        else:
+            for tk, indices in torsion_atoms_dict.items():
+                conf = conf.torsion_energies_one(calculator, 
+                                                indices, 
+                                                simplify, 
+                                                fmax,
+                                                interval, 
+                                                use_converged_only,
+                                                )
+                torsion_energy_profiles[tk] = conf.props['torsion']
         
         self.props['torsion'] = torsion_energy_profiles
         self.props['torsion_calculator'] = str(calculator)
@@ -765,7 +783,7 @@ class Conf:
             return json.dumps(self.props)
     
 
-    def serialize(self, decimals: int = 3, compressed: bool = False) -> str:
+    def serialize(self, decimals: int = 3, compressed: bool = True) -> str:
         """Serialize information necessary to rebuild a Conf object.
         Args:
             decimals (int, optional): number of decimal places for float data type. Defaults to 3.
@@ -788,7 +806,7 @@ class Conf:
         return serialized
 
 
-    def deserialize(self, serialized: str, compressed: bool = False) -> Self:
+    def deserialize(self, serialized: str, compressed: bool = True) -> Self:
         """De-serialize information and rebuild a Conf object.
         
         Example:
@@ -983,3 +1001,89 @@ class Conf:
                           redraw = redraw, 
                           coordgen = coordgen, 
                           optimize = optimize)
+    
+
+def prep_confs_batches(confs: list[Conf], batchsize_atoms: int = 16384) -> list[namedtuple]:
+    """Prepare batches of Chem.Mol conformers for batch optimization.
+
+    - Each batch has up to `batchsize_atoms` number of atoms.
+    - Each batch (namedtuple) has `rdmols`, `size`, `num_atoms` attributes.
+    - Note that this function is not necessary for torsion energies 
+        the batch-optimization of torsion energies are performed using internally prepared batches 
+        of a series of conformers with different dihedral angles.
+
+    Args:
+        batchsize_atoms (int): max number of atoms in one batch. Defaults to 16384(=16*1024).
+
+    Returns:
+        list: list of batches.
+    """
+    Batch = namedtuple('Batch', ['rdmols', 'size', 'num_atoms'])
+    batches = []
+    batch_confs = []
+    batch_atoms = 0
+    for conf in confs:
+        if (batch_atoms + conf.natoms) > batchsize_atoms:
+            batches.append(Batch(rdmols=batch_confs, size=len(batch_confs), num_atoms=batch_atoms))
+            # start over a new batch
+            batch_confs = [ conf.rdmol ]
+            batch_atoms = conf.natoms
+        else:
+            batch_confs.append(conf.rdmol)
+            batch_atoms += conf.natoms
+    if batch_atoms > 0: # last remaining batch
+        batches.append(Batch(rdmols=batch_confs, size=len(batch_confs), num_atoms=batch_atoms))
+
+    return batches
+
+
+def batch_optimize(confs: list[Conf], calculator: Callable, **kwargs) -> list[Conf]:
+    """Batch optimize a list of Conf objects using a given ASE calculator.
+
+    Args:
+        confs (list[Conf]): list of Conf objects to be optimized.
+        calculator (Callable): batch optimizer.
+
+    Returns:
+        list[Conf]: list of optimized Conf objects.
+    """
+    batchsize_atoms = kwargs.get('batchsize_atoms', 16384)
+    batches = prep_confs_batches(confs, batchsize_atoms=batchsize_atoms)
+    optimized_confs = []
+    for batch in batches:
+        optimized = calculator(batch.rdmols, **kwargs).run()
+        for rdmol in optimized.mols:
+            conf = Conf(rdmol)
+            conf.props = {
+                'E_tot_init(kcal/mol)': float(rdmol.GetProp('E_tot_init(kcal/mol)')),
+                'E_tot(kcal/mol)': float(rdmol.GetProp('E_tot(kcal/mol)')),
+                'Converged': True if rdmol.GetProp('Converged') == 'True' else False,
+            }
+            optimized_confs.append(conf)
+    
+    return optimized_confs
+
+
+def batch_singlepoint(confs: list[Conf], calculator: Callable, **kwargs) -> list[Conf]:
+    """Batch singlepoint calculation for a list of Conf objects using a given ASE calculator.
+
+    Args:
+        confs (list[Conf]): list of Conf objects to be calculated.
+        calculator (Callable): batch calculator.
+
+    Returns:
+        list[Conf]: list of Conf objects with updated properties.
+    """
+    batchsize_atoms = kwargs.get('batchsize_atoms', 16384)
+    batches = prep_confs_batches(confs, batchsize_atoms=batchsize_atoms)
+    calculated_confs = []
+    for batch in batches:
+        calculated = calculator(batch.rdmols, **kwargs).run()
+        for rdmol in calculated.mols:
+            conf = Conf(rdmol)
+            conf.props = {
+                'E_tot(kcal/mol)': float(rdmol.GetProp('E_tot(kcal/mol)')),
+            }
+            calculated_confs.append(conf)
+    
+    return calculated_confs
