@@ -7,7 +7,7 @@ import os
 
 from io import StringIO, BytesIO
 from pathlib import Path
-from collections import defaultdict
+from collections import defaultdict, namedtuple
 from collections.abc import Callable
 from typing import Iterator, Self
 
@@ -31,7 +31,7 @@ from rdkit.Chem import (
 from rdkit.ML.Cluster import Butina
 from PIL import Image
 
-from rdworks.conf import Conf
+from rdworks.conf import Conf, batch_optimize, batch_singlepoint
 from rdworks.std import generate_inchi_key, desalt_smiles, standardize, clean_2d
 from rdworks.xml import list_predefined_xml, get_predefined_xml, parse_xml
 from rdworks.scaffold import rigid_fragment_indices
@@ -360,6 +360,22 @@ class Mol:
         return len(unique_rdmols)
    
 
+    def singlepoint(self, calculator: str | Callable, **kwargs) -> Self:
+        """Evaluates potential energy of each conformer without geometry optimization.
+
+        It sets `E_tot(kcal/mol)` property for each conformer.
+
+        Returns:
+            Self: modified self.
+        """
+        if isinstance(calculator, Callable) and 'batchsize_atoms' in kwargs:
+            self.confs = batch_singlepoint(self.confs, calculator, **kwargs)
+        else:
+            for conf in self.confs:
+                PE = conf.singlepoint(calculator, **kwargs) # sets `E_tot(kcal/mol)`
+        return self
+
+
     def make_confs(self, n: int = 50, method: str = 'ETKDG', **kwargs) -> Self:
         """Generates 3D conformers.
 
@@ -451,8 +467,7 @@ class Mol:
 
         # energy evaluations for ranking
         calculator = kwargs.get('calculator', 'MMFF94')
-        for conf in self.confs:
-            conf.potential_energy(calculator) # default: MMFF94
+        self.singlepoint(calculator, **kwargs)
         
         # set relative energy, E_rel(kcal/mol)
         sort_by = 'E_tot(kcal/mol)'
@@ -498,11 +513,16 @@ class Mol:
                     and stable structures.
             fmax (float, optional): fmax for the calculator convergence. Defaults to 0.05.
             max_iter (int, optional): max iterations for the calculator. Defaults to 1000.
+            batchsize_atoms (int, optional): max number of atoms in one batch. 
+                Defaults to 16384(=16*1024). Disable batch optimization if zero or negative.
 
         Returns:
             Self: modified self.
         """
-        self.confs = [ conf.optimize(calculator, fmax, max_iter, **kwargs) for conf in self.confs ]
+        if isinstance(calculator, Callable) and 'batchsize_atoms' in kwargs:
+            self.confs = batch_optimize(self.confs, calculator, **kwargs)
+        else:
+            self.confs = [ conf.optimize(calculator, fmax, max_iter, **kwargs) for conf in self.confs ]
 
         return self
 
@@ -537,9 +557,7 @@ class Mol:
             # re-calculate potential energies
             if verbose :
                 logger.info(f"sort_cons() calculate potential energy by {calculator}")
-
-            for conf in self.confs:
-                PE = conf.potential_energy(calculator, **kwargs) # sets `E_tot(kcal/mol)`
+            self.singlepoint(calculator, **kwargs)
 
         if all(['E_tot(kcal/mol)' in conf.props for conf in self.confs]):
             sort_by = 'E_tot(kcal/mol)'
@@ -918,6 +936,40 @@ class Mol:
         return self
 
 
+    def prep_confs_batches(self, batchsize_atoms: int = 16384) -> list[namedtuple]:
+        """Prepare batches of conformers for batch optimization by neural network potentials (NNP).
+
+        - Each batch has up to `batchsize_atoms` number of atoms.
+        - Each batch (namedtuple) has `rdmols`, `size`, `num_atoms` attributes.
+        - Note that this function is not necessary for torsion energies 
+            the batch-optimization of torsion energies are performed using internally prepared batches 
+            of a series of conformers with different dihedral angles.
+
+        Args:
+            batchsize_atoms (int): max number of atoms in one batch. Defaults to 16384(=16*1024).
+
+        Returns:
+            list: list of batches.
+        """
+        Batch = namedtuple('Batch', ['rdmols', 'size', 'num_atoms'])
+        batches = []
+        batch_confs = []
+        batch_atoms = 0
+        for conf in self.confs:
+            if (batch_atoms + conf.natoms) > batchsize_atoms:
+                batches.append(Batch(rdmols=batch_confs, size=len(batch_confs), num_atoms=batch_atoms))
+                # start over a new batch
+                batch_confs = [ conf.rdmol ]
+                batch_atoms = conf.natoms
+            else:
+                batch_confs.append(conf.rdmol)
+                batch_atoms += conf.natoms
+        if batch_atoms > 0: # last remaining batch
+            batches.append(Batch(rdmols=batch_confs, size=len(batch_confs), num_atoms=batch_atoms))
+
+        return batches
+
+
     def count(self) -> int:
         """Returns the total number of conformers.
 
@@ -939,7 +991,7 @@ class Mol:
         Returns:
             bool: True if applicable.
         """
-        if model.lower() in ['ani-2x', 'ani-2xt']:
+        if model.lower() in ['ani-2x', 'ani-2xt', 'ani2x', 'ani2xt']:
             if self.props['charge'] != 0:
                 return False
             # H, C, N, O, F, S, Cl
@@ -1035,25 +1087,27 @@ class Mol:
             fmax (float, optional): fmax of ASE optimizer. Defaults to 0.05.
             interval (float, optional): interval of torsion angles in degree. Defaults to 15.0.
             use_converged_only (bool, optional): whether to use only converged data. Defaults to True.
-
+            batchsize_atoms (int, optional): max number of atoms in one batch. 
+                Defaults to 16384(=16*1024). Disable batch optimization if zero or negative.
         Returns:
             Self: modified self.
         """
         assert self.count() > 0, "torsion_energies() requires at least one conformer"
         ref_conf = self.confs[0].copy()
         ref_conf = ref_conf.torsion_energies(calculator, 
-                                             torsion_key, 
-                                             simplify, 
-                                             fmax, 
-                                             interval, 
-                                             use_converged_only, 
-                                             **kwargs)
+                                            torsion_key, 
+                                            simplify, 
+                                            fmax, 
+                                            interval, 
+                                            use_converged_only,
+                                            **kwargs)
         for k in ['torsion', 'torsion_calculator']:
             self.props[k] = ref_conf.props[k]
 
         return self
 
 
+
     def similarity(self, other: Self) -> float:
         """Returns Tanimoto similarity with other Mol object.
 
@@ -1601,7 +1655,7 @@ class Mol:
         return json.dumps(props)
 
 
-    def serialize(self, decimals: int = 2, compressed: bool = False) -> str:
+    def serialize(self, decimals: int = 2, compressed: bool = True) -> str:
         """Serialize information necessary to rebuild a Mol object.
 
         Args:
@@ -1613,6 +1667,7 @@ class Mol:
         serialized = json.dumps({
             'name'  : self.name,
             'smiles': self.smiles,
+            'rdmol' : self.to_molblock(),
             'InChIKey': self.InChIKey,
             'props' : recursive_round(self.props, decimals),
             'confs' : [conf.serialize() for conf in self.confs],
@@ -1624,7 +1679,7 @@ class Mol:
         return serialized
 
     
-    def deserialize(self, serialized: str, compressed: bool = False) -> Self:
+    def deserialize(self, serialized: str, compressed: bool = True) -> Self:
         """De-serialize the information and build a new Mol object.
 
         Example:
@@ -1644,15 +1699,25 @@ class Mol:
 
         self.name = data['name']
         self.smiles = data['smiles'] # isomeric SMILES, no H
-        self.rdmol = Chem.MolFromSmiles(data['smiles']) # for 2D depiction
-        self.rdmol.SetProp('_Name', self.name)
+        # self.rdmol = Chem.MolFromSmiles(data['smiles']) # for 2D depiction
+        # self.rdmol.SetProp('_Name', self.name)
+        self.rdmol = Chem.MolFromMolBlock(data['rdmol'])
         self.InChIKey = data['InChIKey']
         self.props = data['props']
         self.confs = [Conf().deserialize(_) for _ in data['confs']] # for 3D conformers (iterable)
         
         return self
     
-    def from_molblock(self, molblock: str) -> Self:
+
+    def to_molblock(self, compressed: bool = False) -> str:
+        """Returns MolBlock"""
+        molblock = Chem.MolToMolBlock(self.rdmol)
+        if compressed:
+            molblock = compress_string(molblock)
+        return molblock
+    
+
+    def from_molblock(self, molblock: str, compressed: bool = False) -> Self:
         """Initialize a new Mol object from MolBlock.
 
         Args:
@@ -1664,7 +1729,12 @@ class Mol:
         Returns:
             Self: self.
         """
+        
+        if compressed:
+            molblock = decompress_string(molblock)
+        
         molecule = Chem.MolFromMolBlock(molblock)
+        
         try:
             self.rdmol, _ = clean_2d(molecule, reset_isotope=True, remove_H=True)
             self.smiles = Chem.MolToSmiles(self.rdmol)
